c************************************************************
c
c               James J. Vincent
c               Kenneth M. Merz
c               Department of Chemistry
c               July, 1994
c
c             ***********************
c             *                     *
c             * Pennsylvania  State *
c             *      University     *
c             *                     *
c             *                     *
c             *    1   8   5   5    *
c             *                     *
c              **                 **
c                ***           ***
c                   ****   ****
c                       ***
c************************************************************
c         Template for  implementation of MPI routines
c     ------------------------------------------------------
c
c   Usage notes:
c
c     Use this template to implement MPI message passing functions
c     on other machines. These functions are required for proper
c     function of parallel AMBER:
c     REDUCE
c     BCAST
c     INIT
c     FINALIZE
c     RANK
c     SIZE
c     BARRIER
c
c***********************************************************

      subroutine MPI_BARRIER(group)
c***********************************************************
c     Synchronize all tasks
**********************************************************
      integer group
      call icc_barrier()
      return
      end

      subroutine MPI_FINALIZE(ierr)
c************************************************************
c     Execute any exit code specific to parallel machine
c************************************************************
      integer ierr
#include "parallel.h"
c     if (master) call kill(0,9)
      stop
      end

      subroutine MPI_INIT(ierr)
c*************************************************************
c     Enrolls caller in parallel machine.
c     Must set number of tasks and own PE number.
c*************************************************************
#include "parallel.h"
      integer ierr

      numtasks=numnodes()
      mytaskid=mynode()
      return
      end

      subroutine MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
c************************************************************
c     Get my PE number (rank)
c************************************************************
#include "parallel.h"
      integer rank,ierr

      rank = mytaskid

      return
      end


      subroutine MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)
c************************************************************
c     Get size of machine (number of PE's)
c************************************************************
#include "parallel.h"
      integer size,ierr

      size = numtasks
      return
      end

      subroutine MPI_RECV(msg,count,datatype,source,tag,comm,ierr)
c***********************************************************
c     Receive a message
c     msg: buffer to put message in
c     count : number of datatype elements
c     datatype: type of data
c     source: PE number of sender
c     tag: message tag
c     comm: group comm id (not used)
c     status: returned status of message receive
c**********************************************************
#include "mpif.h"
c     *** Arguments
      integer count,datatype,source,tag,comm,ierr
      if(datatype.eq.MPI_INTEGER2) then
          call crecvx(tag,msg,count*2,source,-1,msginfo)
      else if(datatype.eq.MPI_INTEGER.or.datatype.eq.MPI_REAL) then
          call crecvx(tag,msg,count*4,source,-1,msginfo)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
          call crecvx(tag,msg,count*8,source,-1,msginfo)
      end if
      return
      end

      subroutine MPI_SEND(msg,count,datatype,dest,tag,comm,ierr)
c***********************************************************
c     Send a message
c     msg: buffer message is in
c     count: number of datatype elements in message
c     datatype MPI datatype of message elements
c     dest: PE number of destination process (0->n-1)
c     tag: message type
c     comm: group identifier for communication (not used yet)
c**********************************************************
#include "mpif.h"
      integer count,datatype,source,tag,comm,ierr
      if(datatype.eq.MPI_INTEGER2) then
          call csend(tag,msg,count*2,dest,0)
      else if(datatype.eq.MPI_INTEGER.or.datatype.eq.MPI_REAL) then
          call csend(tag,msg,count*4,dest,0)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
          call csend(tag,msg,count*8,dest,0)
      end if
      return
      end

      subroutine MPI_BCAST(msg,count,datatype,root,comm,ierr)
c************************************************************
c     Broadcast the array 'msg' to all nodes in the configuration
c     msg type can be anything
c     datatype: type of data being broadcast
c     root: the node which will do the broadcasting
c     group: integer identifier of group to broadcast to -not used
c     count: Number of datatype elements in msg
c************************************************************
#include "mpif.h"

      integer count,datatype,root,comm,ierr,stype
      if(datatype.eq.MPI_INTEGER2) then
         call icc_bcast(msg,count*2,root)
      else if(datatype.eq.MPI_INTEGER.or.datatype.eq.MPI_REAL) then
         call icc_bcast(msg,count*4,root)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
         call icc_bcast(msg,count*8,root)
      end if
      return
      end


      subroutine MPI_REDUCE(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array.
c*****************************************************************
#include "mpif.h"

c     *** Arguments ****
      integer count,datatype,root,comm,ierr
c     This nodes portion of the array to reduce
c     Array to store result in
      if(datatype.eq.MPI_INTEGER2) then
         write(6,*) 'INT*2 global sum not implemented'
         stop
      else if(datatype.eq.MPI_INTEGER) then
         call icc_gisumto1(from,count,to,root)
      else if(datatype.eq.MPI_REAL) then
         call icc_gssumto1(from,count,to,root)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
         call icc_gdsumto1(from,count,to,root)
      end if
      return
      end

      subroutine MPI_ALLREDUCE(from,to,count,datatype,
     &                                   op,comm,ierr)
c************************************************************
c     Perform reduction operation on the array 'from' from all
c     nodes then broadcast the resulting array back to all nodes
c     and place on 'to'.
c************************************************************
#include "mpif.h"

c     *** Arguments
      integer count,datatype,comm,ierr
      external op


      if(datatype.eq.MPI_INTEGER2) then
         write(6,*) 'INT*2 global sum not implemented'
         stop
      else if(datatype.eq.MPI_INTEGER) then
         call icc_gisum(from,count,to)
      else if(datatype.eq.MPI_REAL) then
         call icc_gssum(from,count,to)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
         call icc_gdsum(from,count,to)
      end if
      return
      end

      subroutine MPI_CONCAT(outmsg,inmsg,count,datatype,comm,ierr)
c*********************************************************
c     Concatenate inmsg from each process into outmsg by
c     order of PE #. outmsg is (numtasks)*count in length
c     inmsg: local message to get concatenated
c     outmsg: copy of all messages, placed back on all tasks
c     group: integer identifying group of tasks (not used)
c************************************************************
#include "mpif.h"

c     *** Arguments
      integer count,datatype,comm,ierr
      if(datatype.eq.MPI_INTEGER2) then
         call icc_gcol(outmsg,count*2,inmsg,count*2*numtasks,ndum)
      else if(datatype.eq.MPI_INTEGER.or.datatype.eq.MPI_REAL) then
         call icc_gcol(outmsg,count*4,inmsg,count*4*numtasks,ndum)
      else if(datatype.eq.MPI_DOUBLE_PRECISION) then
         call icc_gcol(outmsg,count*8,inmsg,count*8*numtasks,ndum)
      end if
      return
      end
