#ifndef ANL
c*****************************************************************
c
c     ------------------------------------------------------------
c     PVM implementation of MPI routines for message passing AMBER
c     ------------------------------------------------------------
c
c     The original implementation was written and contributed by:
c                         
c     James J. Vincent    
c     Kenneth M. Merz     
c     Department of Chemistry
c     Pennsylvania State University
c     July, 1994
c
c     COPYRIGHT 1994, Pennsylvania State University.
c                 All rights reserved.  
c
c     This software is covered by the AMBER license 
c     agreement which contains restrictions on its use, 
c     disclosure, and duplication!
c
c     This code basically implements some of the functionality
c     of version 1.0 of MPI (Message Passing Interface) under
c     PVM versions 3.2 and higher.  It has been tested with
c     PVM version 3.3 on cluster of workstations, the IBM
c     SP2, the Cray C90, among other machines.
c
c     To find out more about MPI and/or PVM, 
c     mail netlib@ornl.gov with the text:
c     send index from mpi
c     send index from pvm3
c     ...in the message body.
c
c     USAGE NOTES: Various MACHINE files have been set-up for
c     encountered machines.  If your machine type hasn't been
c     set-up, here is what is necessary.
c
c     Modifications to the MACHINE file...
c
c     (1) Include -DMPI and -DPVM in the MACHINEFLAGS variable for
c         use on machines with standard _unmodified_ PVM3.
c     (2) Add in the appropriate PVM libraries into the LOADLIB 
c         variable, typically:
c         setenv LOADLIB " -lfpvm3 -lgpvm3 -lpvm3 "
c     (3) Add in the path to the PVM libraries into the LOAD
c         variable with the -L option to the compiler, typically
c         -L ~/pvm3/lib/ARCH where ARCH is the name of the
c         PVM architecture for the host you are compiling on.
c     (4) Add in the extra include directory paths to the 
c         CPP variable for the PVM includes with the -I option,
c         typically:
c         setenv CPP "/lib/cpp -I ~/pvm3/include "
c
c     Modifications to the basic source...
c     
c     (*) In the machine specific subdirectory, (Machine/xxx
c         for machine type xxx), modify the bottom of sys.f
c         with the following:
c         #ifdef MPI
c         #include "../mpi/pvm.f"
c         #endif
c
c     IMPORTANT NOTE: If more than one machine type is to be 
c     used in the PVM virtual machine, make sure that the 
c     MACHINE file has turned off the PVM_HOMOGENOUS define so 
c     that PVMDEFAULT will be used instead of PVMRAW.
c
c     ------------------------------------------------------------
c
c     Revisions by tec3:
c     -- modified/added comments
c     -- cleaner include's, requires -I option in MACHINE CPP definition
c     -- added error checking to PVM calls
c
c***********************************************************

      subroutine MPI_BARRIER(MPI_COMM_WORLD,ierr)
c***********************************************************
c     Synchronize pvm tasks with a call to barrier.
c     All pvm tasks are part of 'grp' group - always.
**********************************************************
#include <fpvm3.h>
#include "parallel.h"     

      integer group,info

c     Sync all tasks
      call pvmfbarrier('grp',numtasks,info)
      if (info .lt. 0) then
         write (6, '(''ERROR in MPI_BARRIER()'')')
         call pvmError(info, -1)
      endif
      return
      end

      subroutine MPI_FINALIZE(ierr)
c************************************************************
c     Master task kills off others then exits
c     This will stop the whole program
c     ierr is supposed to be a returned error cndition
c     but here is used as an input exit code
c************************************************************
#include <fpvm3.h>
#include "parallel.h"     

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)
      integer ierr,i
#ifndef CSPP
      if(mytaskid .eq. 0)then
         do i = 0,numtasks-2
            call pvmfkill(ntids(i),ierr)
            if (ierr .lt. 0) then
               write(6, '(''ERROR in MPI_FINALIZE'')')
               call pvmError(info,ntids(i))
            endif
         enddo
      endif
      call pvmfexit(ierr)
#endif
      stop 
      end
      
      subroutine MPI_INIT(ierr)
c*************************************************************
c     Enrolls caller in pvm virtual machine
c     It is assumed all available hosts in the virtual machine
c     will have tasks spawned on them. Nhosts is set by
c     assuming this. Calling task PE# set, and parent pvmtid
c     Parent may not be necessary.
c*************************************************************
#include <fpvm3.h>
#include "parallel.h"     
#include "mpif.h"

c     PVM block local to PVM routines only
      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

c     Returned error code
      integer ierr,i
      character*256 nstring

      integer narch,dtid,speed,info
      character*80 name,arch,hostname
      
c     Enroll this program in PVM 
      call pvmfmytid(mytaskid)
      call pvmfparent(myparent)
#ifdef CSPP
      call pvmfsetopt(PVMPOLLTYPE,PVMPOLLCONSTANT,ierr)
      call getenv('NUMTASKS',nstring)
      read(nstring,*) numtasks
      if (numtasks.lt.1 .or. numtasks.gt.512) then
        write(0,*)' error in NUMTASKS:',numtasks
        call mexit(0,1)
      end if
#else
      call pvmfconfig(numtasks,narch,dtid,name,arch,speed,info)
         if (info .lt. 0) then
            write (6, 991)
            call pvmError(info, -1)
         endif
#endif
c     Everybody joins the group
      call pvmfjoingroup('grp',info)
         if (info .lt. 0) then
            write (6, 992)
            call pvmError(info, -1)
         endif

c     Make master spawn slaves
      if(myparent .EQ. PVMNOPARENT) then
c     If I am master make myparent my own tid
         myparent = mytaskid
         call getarg(0,name)
c     Spawn node programs on all but master
         do i=0,numtasks-2
#ifdef CSPP
           call pvmfspawn(name,PVMDEFAULT,hostname,1,ntids(i),info)
#else
            call pvmfconfig(numtasks,narch,dtid,hostname,
     +           arch,speed,info )
               if (info .lt. 0) then
                  write (6, 991)
                  call pvmError(info, -1)
               endif
            call pvmfspawn(name,PVMHOST,hostname,1,ntids(i),info)
#endif
               if (info .lt. 0) then
                  write (6, 993)
                  call pvmError(info, -1)
               endif
c     Give the newly spawned task its task number 
#ifdef PVM_HOMOGENEOUS
            call pvmfinitsend(PVMRAW,info)
#else 
            call pvmfinitsend(PVMDATADEFAULT,info)
#endif
               if (info .lt. 0) then
                  write (6, 994)
                  call pvmError(info, -1)
               endif
            call pvmfpack(INTEGER4,i+1,1,1,info)
               if (info .lt. 0) then
                  write (6, 995)
                  call pvmError(info, -1)
               endif
            call pvmfsend(ntids(i),96,info)
               if (info .lt. 0) then
                  write (6, 996)
                  call pvmError(info, -1)
               endif
         enddo
         call MPI_BARRIER(MPI_COMM_WORLD,1)
c     All nodes need copy of ntids to reduce from any node
         do i = 0,numtasks-2
#ifdef PVM_HOMOGENEOUS
            call pvmfinitsend(PVMRAW,info)
#else 
            call pvmfinitsend(PVMDATADEFAULT,info)
#endif
               if (info .lt. 0) then
                  write (6, 994)
                  call pvmError(info, -1)
               endif
            call pvmfpack(INTEGER4,ntids(0),numtasks-1,1,info)
               if (info .lt. 0) then
                  write (6, 995)
                  call pvmError(info, -1)
               endif
            call pvmfsend(ntids(i),97,info)
               if (info .lt. 0) then
                  write (6, 996)
                  call pvmError(info, -1)
               endif
         enddo
         mytaskid = 0
      else
         call pvmfrecv(myparent,96,info)
            if (info .lt. 0) then
               write (6, 997)
               call pvmError(info, -1)
            endif
         call pvmfunpack(INTEGER4,mytaskid,1,1,info)
            if (info .lt. 0) then
               write (6, 998)
               call pvmError(info, -1)
            endif
c     Keep a hidden copy of task # for PVM internal stuff
         call MPI_BARRIER(MPI_COMM_WORLD,1)
         call pvmfrecv(myparent,97,info)
            if (info .lt. 0) then
               write (6, 997)
               call pvmError(info, -1)
            endif
         call pvmfunpack(INTEGER4,ntids(0),numtasks-1,1,info)
            if (info .lt. 0) then
               write (6, 998)
               call pvmError(info, -1)
            endif
      endif
      mtids(0)=myparent
      do i = 1,numtasks-1
         mtids(i)=ntids(i-1)
      enddo

 991  format('Error in MPI_INIT(), call to pvmfconfig()')
 992  format('Error in MPI_INIT(), call to pvmfjoingroup()')
 993  format('Error in MPI_INIT(), call to pvmfspawn()')
 994  format('Error in MPI_INIT(), call to pvmfinitsend()')
 995  format('Error in MPI_INIT(), call to pvmfpack()')
 996  format('Error in MPI_INIT(), call to pvmfsend()')
 997  format('Error in MPI_INIT(), call to pvmfrecv()')
 998  format('Error in MPI_INIT(), call to pvmfunpack()')
      
      return
      end

      subroutine MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
c************************************************************
c     Get my PE number (rank) 
c************************************************************
#include "parallel.h"     

      integer rank,ierr

      rank = mytaskid
      return
      end


      subroutine MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)
c************************************************************
c     Get size of virtual machine (number of PE's)
c************************************************************
#include "parallel.h"     

      integer size,ierr

      size = numtasks
      return
      end

      subroutine MPI_RECV(msg,count,datatype,source,tag,comm,ist,ierr)
c***********************************************************
c     Receive a message 
c     msg: buffer to put message in
c     count : number of datatype elements
c     datatype: type of data
c     source: PE number of sender
c     tag: message tag
c     comm: group comm id (not used)
c     ist: returned status of message receive, not used
c**********************************************************
#include <fpvm3.h>
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

c     *** Arguments 
      dimension msg(*)
      double precision msg
      integer count,datatype,source,tag,comm,ierr
c     *** Local
      integer stid

      if(source.eq.0)then
         stid = myparent
      else
c     This should probably be for source == MPI_ALL or something
         if (source .eq. -1)then
            stid = -1
         else
c     Only node tid's are in ntids() and start at 0
            stid = ntids(source-1)
         endif
      endif
      call pvmfrecv(stid,tag,ierr)
         if (ierr .lt. 0) then
            write (6, '(''Error in MPI_RECV(), call to pvmfrecv'')')
            call pvmError(ierr, -1)
         endif
      call pvmfunpack(datatype,msg,count,1,ierr)
         if (ierr .lt. 0) then
            write (6, '(''Error in MPI_RECV(), call to pvmfunpack'')')
            call pvmError(ierr, -1)
         endif

      return
      end

      subroutine MPI_SEND(msg,count,datatype,dest,tag,comm,ierr)
c***********************************************************
c     Send a message 
c     msg: buffer message is in
c     count: number of datatype elements in message
c     datatype MPI datatype of message elements
c     dest: PE number of destination process (0->n-1)
c     tag: message type
c     comm: group identifier for communication (not used yet)
c**********************************************************
#include <fpvm3.h>
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

      dimension msg(*)
      double precision msg

      integer count,datatype,dest,tag,comm,ierr
      integer stype,dtid

      if(dest.eq.0) then
         dtid = myparent
      else
         dtid = ntids(dest-1)
      endif
#ifdef PVM_HOMOGENEOUS
      call pvmfinitsend(PVMRAW,info)
#else 
      call pvmfinitsend(PVMDATADEFAULT,info)
#endif
         if (info .lt. 0) then
            write (6, 
     .           '(''Error in MPI_SEND(), call to pvmfinitsend'')')
            call pvmError(info, -1)
         endif
      call pvmfpack(datatype,msg,count,1,info)
         if (info.lt. 0) then
            write (6, '(''Error in MPI_RECV(), call to pvmfpack'')')
            call pvmError(info, -1)
         endif
      call pvmfsend(dtid,tag,info)
         if (info .lt. 0) then
            write (6, '(''Error in MPI_RECV(), call to pvmfsend'')')
            call pvmError(info, -1)
         endif

      return
      end

      subroutine MPI_BCAST(msg,count,datatype,root,comm,ierr)
c************************************************************
c     Broadcast the array 'msg' to all nodes in the configuration
c     (should be all nodes in group really)
c     msg type can be anything
c     datatype: type of data being broadcast
c     root: the node which will do the broadcasting
c     group: integer identifier of group to broadcast to -not used
c     count: Number of datatype elements in msg
c     **** PVM msgtype 98 is used for the communication *****
c************************************************************
#include <fpvm3.h>
#include "parallel.h"     
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

      integer count,datatype,root,comm,ierr,stype
      
c     This nodes array 
      dimension msg(*)
      double precision msg

      if(mytaskid.eq.root) then
#ifdef PVM_HOMOGENEOUS
            call pvmfinitsend(PVMRAW,info)
#else 
            call pvmfinitsend(PVMDATADEFAULT,info)
#endif
            if (info.lt. 0) then
               write (6, 
     .              '(''Error in MPI_RECV(), call to pvmfinitsend'')')
               call pvmError(info, -1)
            endif
         call pvmfpack(datatype,msg,count,1,info)
            if (info .lt. 0) then
               write (6, '(''Error in MPI_RECV(), call to pvmfpack'')')
               call pvmError(info, -1)
            endif
	if(numtasks.GT.1) then
         call pvmfbcast('grp',98,info)
	endif
            if (info .lt. 0) then
               write (6, 
     .              '(''Error in MPI_RECV(), call to pvmfbcast'')')
               call pvmError(info, -1)
            endif
      else
         call MPI_RECV(msg,count,datatype,root,98,comm,ist,info)         
      endif
      
      return 
      end
      
      subroutine MPI_REDUCE(from,to,count,datatype,op,root,comm,ierr)
c
c  --wrapper to check datatype and call correct routines:
c
#include "mpif.h"
      integer datatype
c
      if (datatype.eq.MPI_INTEGER) then
        call MPI_REDUCE_I(from,to,count,datatype,op,root,comm,ierr)
      else if (datatype.eq.MPI_DOUBLE_PRECISION) then
        call MPI_REDUCE_D(from,to,count,datatype,op,root,comm,ierr)
      else
        write(0,*) 'error in mpi-reduce call:',datatype
        call mexit(6,1)
      end if
      return
      end
c
      subroutine MPI_REDUCE_I(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array. Reduction operation is
c     double precision add only. Result is placed on 'node'.
c     *** PVM message type 99 used for this operation ****
c     'from' array on root node gets overwritten with garbage
c*****************************************************************
#include <fpvm3.h>
#include "parallel.h"     
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

c     *** Arguments ****
      integer count,datatype,root,comm,ierr
c     external op
c
c     This nodes portion of the array to reduce:
c
      integer from
      dimension from(*)
c
c     Array to store result in:
c
      integer to
      dimension to(*)

c
c     *** Local     ****
c
      integer i,j,nodecnt,stype

      call MPI_BARRIER(MPI_COMM_WORLD,1)
c
c     If I am root then I am the receiver
c
      if(root .eq. mytaskid) then      
c
c     Add in my portion of array
c
         do i = 1,count
            to(i) = from(i)
         enddo
         do i = 1,numtasks-1
c
c     Get copy of 'from' array from all other nodes 
c
            call MPI_RECV(from,count,datatype,-1,99,comm,ist,ierr)
c
c     Add in new copy of array
c
            do j = 1,count
               to(j) = to(j) + from(j)
            enddo
         enddo
c
c     If I am not the receving node send my data off
c
      else
         call MPI_SEND(from,count,datatype,root,99,comm,ierr)
       endif

      return
      end
c
      subroutine MPI_REDUCE_D(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array. Reduction operation is
c     double precision add only. Result is placed on 'node'.
c     *** PVM message type 99 used for this operation ****
c     'from' array on root node gets overwritten with garbage
c*****************************************************************
#include <fpvm3.h>
#include "parallel.h"     
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

c     *** Arguments ****
      integer count,datatype,root,comm,ierr
c     external op
c     This nodes portion of the array to reduce
      dimension from(*)
      double precision from
c     Array to store result in
      dimension to(*)
      double precision to

c     *** Local     ****
c     Counters 
      integer i,j,nodecnt,stype
      double precision temp

      call MPI_BARRIER(MPI_COMM_WORLD,1)
c     If I am root then I am the receiver
      if(root .eq. mytaskid) then      
c     Add in my portion of array
         do i = 1,count
            to(i) = from(i)
         enddo
         do i = 1,numtasks-1
c     Get copy of 'from' array from all other nodes 
            call MPI_RECV(from,count,datatype,-1,99,comm,ist,ierr)

c     Add in new copy of array
            do j = 1,count
               to(j) = to(j) + from(j)
            enddo
         enddo
c     If I am not the receving node send my data off
      else
         call MPI_SEND(from,count,datatype,root,99,comm,ierr)
       endif

      return
      end

      subroutine MPI_ALLREDUCE(from,to,count,datatype,
     &                                   op,comm,ierr)
c************************************************************
c     Perform reduction operation on the array 'from' from all
c     nodes then broadcast the resulting array back to all nodes
c     and place on 'to'. 
c     Only add supported now, but others are easily added
c************************************************************
#include <fpvm3.h>

c     *** Argumenst
      dimension from(*),to(*)
      double precision to,from
      integer count,datatype,comm,ierr
c     external op
#ifdef DISTSUM
#include "parallel.h"
#include "mpif.h"
      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)
      dimension jparpt(0:512)
c
      if (datatype.eq.MPI_DOUBLE_PRECISION) then
        call pvmfgblsum(mytaskid,mtids,numtasks,204,from,to,count,
     .      jparpt,ierr)
        do i=1,count
          to(i) = from(i)
        end do
      else
c       Use node 0 as the central gatherer
        call MPI_REDUCE(from,to,count,datatype,op,0,comm,ierr)
        call MPI_BCAST(to,count,datatype,0,comm,ierr)
      end if
#else
c     Use node 0 as the central gatherer
      call MPI_REDUCE(from,to,count,datatype,op,0,comm,ierr)
      call MPI_BCAST(to,count,datatype,0,comm,ierr)
#endif
      return
      end

      subroutine MPI_CONCAT(outmsg,inmsg,count,datatype,comm,ierr)
c*********************************************************
c     Concatenate outmsg from each process into inmsg by
c     order of PE #. Inmsg is (nhosts)*outmsg in length
c     outmsg: local message to get concatenated
c     inmsg: copy of all messages, placed back on all tasks
c     size: size of OUTMSG in BYTES
c     group: integer identifying group of tasks (not used)
c     MSGTYPE 102 used here
c     !!!!outmsg gets destroyed in the process!!!
c************************************************************
#include <fpvm3.h>
#include "parallel.h"     
#include "mpif.h"

      common/locpvm/ntids(0:511),myparent,mynode,mtids(0:511)

c     *** Arguments
      dimension outmsg(*),inmsg(*)
      integer outmsg,inmsg
      integer count,datatype,comm,ierr

c     *** Local 
      integer i

      call MPI_BARRIER(MPI_COMM_WORLD,1)

c     Everyone send their data to 0 (including 0)
      call MPI_SEND(outmsg,count,datatype,0,102,comm,ierr)

c     PE # 0 does the work
      if(mytaskid .eq. 0) then      
c     Make sure to recv messages in order
         do i = 0,numtasks-1
            call MPI_RECV(inmsg(i*count+1),count,datatype,
     +                    i,102,comm,ist,ierr)
         enddo
      endif
c     Broadcast concatenated msg back to all
      call MPI_BCAST(inmsg,count*numtasks,datatype,0,comm,ierr)
      return
      end

c     PVMERROR() 
c     ...a routine which prints out the type of PVM error code
c     based on PVM 3.3 error codes.  
c     NOTE: these errors currently are not fatal; 
c     perhaps they should be? (tec3)
c
      subroutine pvmError(ierr, itid)
#include <fpvm3.h>
      integer ierr, itid

c     This routine is based on the PVM errors as listed in
c     PVM version 3.3

      if (itid .ne. -1) then
         write(6, '(''Task ID is '', i4)') itid
      endif
      if (ierr .eq. -2) then
         write(6, '(''Error is '', a20)') 'PvmBadParam'
      elseif (ierr .eq. -3) then
         write(6, '(''Error is '', a20)') 'PvmMismatch'
      elseif (ierr .eq. -5) then
         write(6, '(''Error is '', a20)') 'PvmNoData'
      elseif (ierr .eq. -6) then
         write(6, '(''Error is '', a20)') 'PvmNoHost'
      elseif (ierr .eq. -7) then
         write(6, '(''Error is '', a20)') 'PvmNoFile'
      elseif (ierr .eq. -10) then
         write(6, '(''Error is '', a20)') 'PvmNoMem'
      elseif (ierr .eq. -12) then
         write(6, '(''Error is '', a20)') 'PvmBadMsg'
      elseif (ierr .eq. -14) then
         write(6, '(''Error is '', a20)') 'PvmSysErr'
      elseif (ierr .eq. -15) then
         write(6, '(''Error is '', a20)') 'PvmNoBuf'
      elseif (ierr .eq. -16) then
         write(6, '(''Error is '', a20)') 'PvmNoSuchBuf'
      elseif (ierr .eq. -17) then
         write(6, '(''Error is '', a20)') 'PvmNoSuchBuf'
      elseif (ierr .eq. -18) then
         write(6, '(''Error is '', a20)') 'PvmDupGroup'
      elseif (ierr .eq. -19) then
         write(6, '(''Error is '', a20)') 'PvmNoGroup'
      elseif (ierr .eq. -20) then
         write(6, '(''Error is '', a20)') 'PvmNotInGroup'
      elseif (ierr .eq. -21) then
         write(6, '(''Error is '', a20)') 'PvmNoInst'
      elseif (ierr .eq. -22) then
         write(6, '(''Error is '', a20)') 'PvmHostFail'
      elseif (ierr .eq. -23) then
         write(6, '(''Error is '', a20)') 'PvmNoParent'
      elseif (ierr .eq. -24) then
         write(6, '(''Error is '', a20)') 'PvmNotImpl'
      elseif (ierr .eq. -25) then
         write(6, '(''Error is '', a20)') 'PvmDSysErr'
      elseif (ierr .eq. -26) then
         write(6, '(''Error is '', a20)') 'PvmBadVersion'
      elseif (ierr .eq. -27) then
         write(6, '(''Error is '', a20)') 'PvmOutOfRes'
      elseif (ierr .eq. -28) then
         write(6, '(''Error is '', a20)') 'PvmDupHost'
      elseif (ierr .eq. -29) then
         write(6, '(''Error is '', a20)') 'PvmCantStart'
      elseif (ierr .eq. -30) then
         write(6, '(''Error is '', a20)') 'PvmAlready'
      elseif (ierr .eq. -31) then
         write(6, '(''Error is '', a20)') 'PvmNoTask'
      elseif (ierr .eq. -32) then
         write(6, '(''Error is '', a20)') 'PvmNoEntry'
      elseif (ierr .eq. -33) then
         write(6, '(''Error is '', a20)') 'PvmDupEntry'
      endif
#ifdef PVM_ABORT_ON_ERROR
      call MPI_FINALIZE(ierr)
#endif
      return
      end

#endif
#ifdef CSPP
      subroutine pvmfgdsum(node,tids,nproc,msgtype,x,temp,iparpt,err)
      implicit none
#include <fpvm3.h>
#ifdef PVMTIM
#  include "pvmprof.h"
#endif
      integer node,nproc,tids(0:nproc-1)
      integer msgtype,logtwo
      real*8 x(0:*),temp(0:*)
      integer iparpt(0:nproc)
      logical err
C---
      integer other,ncycles,k,bit,cs,cr,i
      integer ns,nr,status,bufid

      err=.false.
      if (nproc.le.1) return

      ncycles = logtwo(nproc)
      bit=nproc/2
      do k = 0,ncycles-1
         other=ieor(node,bit)
C        send chunk
         cs = ishft(other,-((ncycles-1)-k))*bit
C        recv chunk
         cr = ishft(node,-((ncycles-1)-k))*bit
C        pack send buffer
         ns = iparpt(cs+bit)-iparpt(cs)
         if (ns.gt.0) then
#if PVMTIM
            packtimr=walltime(0.0D0)
#endif
            call pvmfinitsend(PVMRAW,bufid)
            call pvmfpack(REAL8,x(iparpt(cs)),ns,1,status)
            if(status.ne.PvmOk) then
               err = .true.
               return
            endif
#if PVMTIM
            packtim=walltime(packtimr)
            packtima=packtima+packtim
#endif
#if PVMTIM
            sndtimr=walltime(0.0D0)
#endif
            call pvmfsend(tids(other),msgtype,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
#if PVMTIM
            sndtim=walltime(sndtimr)
            sndtima=sndtima+sndtim
#endif
         endif
         nr = iparpt(cr+bit)-iparpt(cr)
         if(nr.gt.0) then
#if PVMTIM
            rcvtimr=walltime(0.0D0)
#endif
            call pvmfrecv(tids(other),msgtype,status)
            if(status.lt.0) then
               err=.true.
               return
            endif
#if PVMTIM
            rcvtim=walltime(rcvtimr)
            rcvtima=rcvtima+rcvtim
#endif
#if PVMTIM
            unpacktimr=walltime(0.0D0)
#endif
            call pvmfunpack(REAL8,temp(iparpt(cr)),nr,1,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
#if PVMTIM
            unpacktim=walltime(unpacktimr)
            unpacktima=unpacktima+unpacktim
#endif
            do i=iparpt(cr),iparpt(cr+bit)-1
                 x(i)=x(i)+temp(i)
            enddo
         endif
         bit = ishft(bit,-1)
      enddo
      return
      end
      subroutine pvmfdbcast(node,tids,nproc,msgtype,x,iparpt,err)
      implicit none
#include <fpvm3.h>
#if PVMTIM
#include "pvmprof.h"
#endif
      integer node,nproc,tids(0:nproc-1)
      integer msgtype,logtwo
      real*8 x(0:*)
      integer iparpt(0:nproc)
      logical err
C---
      integer other,ncycles,k,bit,cs,cr
      integer ns,nr,status,bufid

      err=.false.
      if (nproc.le.1) return

      ncycles = logtwo(nproc)
      bit=1
      do k = 0,ncycles-1
         other=ieor(node,bit)
C        send chunk
         cs = ishft(node,-k)*bit
C        recv chunk
         cr = ishft(other,-k)*bit
C        pack send buffer
         ns = iparpt(cs+bit)-iparpt(cs)
         if (ns.gt.0) then
#if PVMTIM
            packtimr=walltime(0.0D0)
#endif
            call pvmfinitsend(PVMRAW,bufid)
            call pvmfpack(REAL8,x(iparpt(cs)),ns,1,status)
            if(status.ne.PvmOk) then
               err = .true.
               return
            endif
#if PVMTIM
            packtim=walltime(packtimr)
            packtima=packtima+packtim
#endif
         endif
         if(ns.gt.0) then
#if PVMTIM
            sndtimr=walltime(0.0D0)
#endif
            call pvmfsend(tids(other),msgtype,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
#if PVMTIM
            sndtim=walltime(sndtimr)
            sndtima=sndtima+sndtim
#endif
         endif
         nr = iparpt(cr+bit)-iparpt(cr)
         if(nr.gt.0) then
#if PVMTIM
            rcvtimr=walltime(0.0D0)
#endif
            call pvmfrecv(tids(other),msgtype,status)
            if(status.lt.0) then
               err=.true.
               return
            endif
#if PVMTIM
            rcvtim=walltime(rcvtimr)
            rcvtima=rcvtima+rcvtim
#endif
#if PVMTIM
            unpacktimr=walltime(0.0D0)
#endif
            call pvmfunpack(REAL8,x(iparpt(cr)),nr,1,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
#if PVMTIM
            unpacktim=walltime(unpacktimr)
            unpacktima=unpacktima+unpacktim
#endif
         endif
         bit = ishft(bit,1)
      enddo
      return
9999  err=.true.
      return
      end

#else
      subroutine pvmfgdsum(node,tids,nproc,msgtype,x,temp,iparpt,err)
      implicit none
#include <fpvm3.h>
      integer node,nproc,tids(0:nproc-1)
      integer msgtype,logtwo
      real*8 x(0:*),temp(0:*)
      integer iparpt(0:nproc)
      logical err
C---
      integer other,ncycles,k,bit,cs,cr,i
      integer ns,nr,status,bufid
      logical sendfirst

      err=.false.
      if (nproc.le.1) return

      ncycles = logtwo(nproc)
      bit=nproc/2
      do k = 0,ncycles-1
         other=ieor(node,bit)
C        send chunk
         cs = ishft(other,-((ncycles-1)-k))*bit
C        recv chunk
         cr = ishft(node,-((ncycles-1)-k))*bit
C        pack send buffer
         ns = iparpt(cs+bit)-iparpt(cs)
         if (ns.gt.0) then
            call pvmfinitsend(PVMRAW,bufid)
            call pvmfpack(REAL8,x(iparpt(cs)),ns,1,status)
            if(status.ne.PvmOk) then
               err = .true.
               return
            endif
         endif
C---     do we send now or do the receive first?
         sendfirst = iand(node,bit).gt.0
         if (sendfirst) then
            if(ns.gt.0) then
               call pvmfsend(tids(other),msgtype,status)
               if(status.ne.PvmOk) then
                  err=.true.
                  return
               endif
            endif
         endif
         nr = iparpt(cr+bit)-iparpt(cr)
         if(nr.gt.0) then
            call pvmfrecv(tids(other),msgtype,status)
            if(status.lt.0) then
               err=.true.
               return
            endif
            call pvmfunpack(REAL8,temp(iparpt(cr)),nr,1,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
            do i=iparpt(cr),iparpt(cr+bit)-1
                 x(i)=x(i)+temp(i)
            enddo
         endif
         if (.not.sendfirst) then
            if(ns.gt.0) then
               call pvmfsend(tids(other),msgtype,status)
               if(status.ne.PvmOk) then
                  err=.true.
                  return
               endif
            endif
         endif
         bit = ishft(bit,-1)
      enddo
      return
      end
      subroutine pvmfdbcast(node,tids,nproc,msgtype,x,iparpt,err)
      implicit none
#include <fpvm3.h>
      integer node,nproc,tids(0:nproc-1)
      integer msgtype,logtwo
      real*8 x(0:*)
      integer iparpt(0:nproc)
      logical err
C---
      integer other,ncycles,k,bit,cs,cr
      integer ns,nr,status,bufid
      logical sendfirst

      err=.false.
      if (nproc.le.1) return

      ncycles = logtwo(nproc)
      bit=1
      do k = 0,ncycles-1
         other=ieor(node,bit)
C        send chunk
         cs = ishft(node,-k)*bit
C        recv chunk
         cr = ishft(other,-k)*bit
C        pack send buffer
         ns = iparpt(cs+bit)-iparpt(cs)
         if (ns.gt.0) then
            call pvmfinitsend(PVMRAW,bufid)
            call pvmfpack(REAL8,x(iparpt(cs)),ns,1,status)
            if(status.ne.PvmOk) then
               err = .true.
               return
            endif
         endif
C---     do we send now or do the receive first?
         sendfirst = iand(node,bit).gt.0
         if (sendfirst) then
            if(ns.gt.0) then
               call pvmfsend(tids(other),msgtype,status)
               if(status.ne.PvmOk) then
                  err=.true.
                  return
               endif
            endif
         endif
         nr = iparpt(cr+bit)-iparpt(cr)
         if(nr.gt.0) then
            call pvmfrecv(tids(other),msgtype,status)
            if(status.lt.0) then
               err=.true.
               return
            endif
            call pvmfunpack(REAL8,x(iparpt(cr)),nr,1,status)
            if(status.ne.PvmOk) then
               err=.true.
               return
            endif
         endif
         if (.not.sendfirst) then
            if(ns.gt.0) then
               call pvmfsend(tids(other),msgtype,status)
               if(status.ne.PvmOk) then
                  err=.true.
                  return
               endif
            endif
         endif
         bit = ishft(bit,1)
      enddo
      return
9999  err=.true.
      return
      end
#endif
      subroutine pvmfgblsum(node,tids,nproc,msgtype,x,temp,
     .      nval,jparpt,err)
      implicit none
#include <fpvm3.h>
      integer nproc,nchunk,nremnd
      integer tids(0:nproc-1),node,msgtype,nval
      integer jparpt(0:nproc)
      real*8 x(0:nval-1),temp(0:nval-1)
      logical err

      integer i
      
      err=.false.
      if (nproc.le.1) return
      if (nval.lt.nproc) then
        jparpt(0)=0
        do i=1,nproc
          jparpt(i)=nval 
        enddo
      else
C       Figure out the partioning of the atoms
        JPARPT(0) = 0
        NCHUNK = nval/nproc
        NREMND = nval - NCHUNK*nproc
        IF (NREMND.GT.0) THEN
           DO I = 1,NREMND
              JPARPT(I) = JPARPT(I-1) + NCHUNK + 1
           ENDDO
           DO I = NREMND+1,nproc
              JPARPT(I) = JPARPT(I-1) + NCHUNK
           ENDDO
        ELSE
           DO I = 1,nproc
              JPARPT(I) = JPARPT(I-1) + NCHUNK
           ENDDO
        ENDIF
      endif
      call pvmfgdsum(node,tids,nproc,msgtype,x,temp,jparpt,err)
      if(err) return
      call pvmfdbcast(node,tids,nproc,msgtype,x,jparpt,err)
      return
      end
