c
c     Primitive MPI wrappers for the Cray T3E using the Cray
c     shmem library functionality.  Redesigned for AMBER 5.0
c     to allow specification of "communicators" and to cycle
c     between scratch sychronization arrays
c
c     Developed by Michael Crowley 
c     (Pittsburgh Supercomputing Center)
c

c
c------------------------------------------------------------
c
c     MPI_BARRIER
c
c------------------------------------------------------------
c
      subroutine MPI_BARRIER(comm,rcode)

      implicit none
#     include "t3e.h"
#     include "parallel.h"
      integer rcode,shmem_n_pes,shmem_my_pe,comm

      call shmem_barrier(mpes(comm,1),0,mpes(comm,2),
     $     barsync(1,bar_cnt(comm),comm))
      bar_cnt(comm)=mod(bar_cnt(comm),T3E_BARRSYNC_SZ)+1
      return
      end

c------------------------------------------------------------
c
c     MPI_ABORT
c
c------------------------------------------------------------

      subroutine mpi_abort(comm, errorcode, ierror)
      implicit none
      integer comm, errorcode, ierror

#include "mpif.h"
#include "parallel.h"

#ifdef MPI_VERBOSE
      write (6,*) 'MPI_ABORT called with errorcode ', errorcode,
     .     ' processor ', mytaskid
#endif
      call globalexit(errorcode)

      return
      end

c------------------------------------------------------------
c
c     MPI_FINALIZE
c
c------------------------------------------------------------

      subroutine MPI_FINALIZE(rcode)

      implicit none
      integer rcode
      call exit(rcode)
      return
      end
      

c------------------------------------------------------------
c
c     MPI_INIT
c
c------------------------------------------------------------

      subroutine MPI_INIT(err)
      implicit none

#include "t3e.h"
#include "parallel.h"
#include "mpif.h"

      integer sem1,semwait,semset
      common/semaphores/sem1,semwait,semset

      integer err
      integer i,j,k

      integer shmem_n_pes,shmem_my_pe
c
c     Initialize syncronization vectors
c
      do j = 1,T3E_DATASYNC_SZ
         do i = 1,SHMEM_REDUCE_SYNC_SIZE
            datasync(i,j) = SHMEM_SYNC_VALUE
         enddo
      enddo
      data_cnt = 1

      do k = 1, TOTAL_MPI_COMMUNICATORS
         do j = 1,T3E_BARRSYNC_SZ
            do i = 1,SHMEM_BARRIER_SYNC_SIZE
               barsync(i,j,k) = SHMEM_SYNC_VALUE
            enddo
         enddo
         bar_cnt(k)=1
      enddo

      do j = 1,T3E_BCSYNC_SZ
         do i = 1,SHMEM_BCAST_SYNC_SIZE
            bcastsync(i,j) = SHMEM_SYNC_VALUE
         enddo
      enddo
      bcast_cnt = 1

      do j = 1,T3E_CLSYNC_SZ
         do i = 1,SHMEM_COLLECT_SYNC_SIZE
            collsync(i,j) = SHMEM_SYNC_VALUE
         enddo
      enddo
      coll_cnt = 1


      mytaskid = shmem_my_pe()
      numtasks = shmem_n_pes()
      do i=1,TOTAL_MPI_COMMUNICATORS
         mpi_numgp_active(i)= MPI_GROUP_NULL
         mpi_ncomm_active(i)= MPI_COMM_NULL
         mpi_gpcommnum(i)=MPI_GROUP_NULL
      enddo
                                                   
      mpi_numgp_active(0)=1
      mpi_ncomm_active(0)=0
      mpi_gpsize(0)=numtasks
      mpi_gpcommnum(0)=0
      mpi_cmsize(0)=numtasks
      mpi_cmrank0(0)=0
      do i=0,numtasks-1
         mpi_gpranks(i,0)=i
      enddo
      mpes(0,1) = 0
      mpes(0,2) = numtasks
      
      mygroupid = MPI_COMM_WORLD
      grptasks = numtasks

#ifndef T3E
      call shmem_set_cache_inv()
#endif

c     JV Intialize semaphore
      semset = 1
      semwait = -99
      sem1 = semwait

      call barrier()

      return
      end

c************************************************************
c               MPI_COMM_RANK
c************************************************************
      subroutine MPI_COMM_RANK(comm,rank,ierror)
      implicit none
#include "t3e.h"
#include "parallel.h"
#include "mpif.h"

      integer comm,rank,ierror, shmem_my_pe
      integer i, sh_rank, group

      sh_rank = shmem_my_pe()
      
      group = mpi_gpcommnum(comm)
      rank = MPI_UNDEFINED
      do i=0,mpi_gpsize(group)-1
         if(mpi_gpranks(i,group).eq.sh_rank)rank = i
      enddo

      return
      end


c************************************************************
c               MPI_ALLGATHERV
c************************************************************
      subroutine MPI_ALLGATHERV(
     $     sendbuf, sendcount, sendtype,
     $     recvbuf, recvcounts, displs, recvtype, 
     $     comm,ierror)
      implicit none
      real sendbuf(*), recvbuf(*)
      integer sendcount, sendtype,  displs(*),
     $     recvcounts(*),recvtype, 
     $     comm, ierror
      integer i
#include "t3e.h"
#include "parallel.h"
#include "mpif.h"

#ifdef MPI_VERBOSE
      if(mytaskid.eq.0) write(6,*) 'MPI_ALLGATHERV'
#endif


# ifdef MPI_VERBOSE
      if(mytaskid.eq.0) write(6,*) '....... shmem_collect',
     $     (recvcounts(i),i=1,numtasks)
# endif

      call shmem_collect(recvbuf,sendbuf,recvcounts(mytaskid+1),
     $ 0,0,numtasks, collsync(1,coll_cnt))   
            coll_cnt = mod(coll_cnt,T3E_CLSYNC_SZ) + 1
      call MPI_BARRIER(MPI_COMM_WORLD,1)


      return
      end


      subroutine MPI_GATHER(sendbuf, sendcount, sendtype,
     .   recvbuf, recvcounts, recvtype, root, comm,
     .   ierror)
      double precision sendbuf(*), recvbuf(*)
      integer sendcount, sendtype, recvcounts(*),
     .   recvtype, root, comm, ierror

#include "t3e.h"
#include "parallel.h"
#include "mpif.h"

      call shmem_collect(recvbuf,sendbuf,sendcount,
     $ 0,0,numtasks, collsync(1,coll_cnt))   
            coll_cnt = mod(coll_cnt,T3E_CLSYNC_SZ) + 1
      call MPI_BARRIER(MPI_COMM_WORLD,1)

      return
      end




c************************************************************
c               MPI_REDUCE_SCATTER
c************************************************************
      subroutine MPI_REDUCE_SCATTER(sendbuf, recvbuf, recvcounts,
     .   datatype, op, comm, ierror)
      implicit none
      real sendbuf(*), recvbuf(*)
      integer recvcounts(*), datatype, op, comm, ierror
#include "parallel.h"
#include "t3e.h"
#include "mpif.h"

      integer count, offset
      integer i
#ifdef MPI_VERBOSE
      if(mytaskid.eq.0) write(6,*) 'MPI_REDUCE_SCATTER',count
#endif
      count=0
      do i=1,numtasks
         count=count+recvcounts(i)
      enddo

      offset = 0
      do i = 1, mytaskid
         offset = offset + recvcounts(i)
      enddo
      call MPI_BARRIER(MPI_COMM_WORLD, 1)

      if(count/2+1.gt.T3E_WRKDATA_SIZE)then
         write (6,*)
     $        'T3E_WRKDATA_SIZE too small (', T3E_WRKDATA_SIZE,')'
         write (6,*)
     $        ' MPI_REDUCE_SCATTER summing ',count,' elements'
         write (6,*)
     $        '      needs wrkdata at least ',count/2+1
         write (6,*)
     $        '   redimension in t3e.h '
         call mexit(99)
      endif

      call shmem_real8_sum_to_all(
     .     recvbuf(1-offset),
     .     sendbuf,
     .     count,
     .     0,
     .     0,
     .     numtasks,
     .     wrkdata,
     .     datasync(1,data_cnt))
      data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1

      call MPI_BARRIER(MPI_COMM_WORLD, 1)

      return
      end

c************************************************************
c               MPI_RECV
c************************************************************
      subroutine MPI_RECV(msg,ncnt,ntype,nsrc,ntag,ncomm,ist,err)
c***********************************************************
c     Receive a message 
c     msg: buffer to return message in
c     count : number of datatype elements
c     datatype: type of data - not used
c     source: PE number of sender
c     tag: message tag - not used
c     comm: group comm id - not used
c     status: returned status of message receive - not used
c     err:    returned error condition - not used
c**********************************************************
      implicit none
#include "t3e.h"
      integer sem1,semwait,semset
      common/semaphores/sem1,semwait,semset
c     *** Arguments 
      integer msg(*)
c      integer count,datatype,source,tag,comm,err,i
      integer ncnt,ntype,nsrc,ntag,ncomm,ist,err, i

#include "parallel.h"

c     Set semaphore on source PE
      call shmem_fence
      call shmem_put(sem1,semset,1,nsrc)   
c     JV Wait on semaphore until msg is put on this node
      if(sem1 .EQ. semwait)call shmem_wait(sem1,semwait)
      sem1 = semwait
c     JV Move message from buffer to correct place
      do i = 1,ncnt
         msg(i) = nbuf(i)
      enddo
#ifdef MPI_VERBOSE
      write (6,*) 'PE',mytaskid,' MPI_RECV',nsrc
#endif
      return
      end

c************************************************************
c               MPI_SEND
c************************************************************
      subroutine MPI_SEND(msg,count,datatype,dest,tag,comm,err)
c***********************************************************
c     Send a message 
c     msg: buffer message is in
c     count: Number of datatype elements in msg
c     datatype: data type of msg elements - not used since all 
c               datatypes are same size on T3E
c     dest: PE number of destination process (0->n-1)
c     tag: message ID - not used
c     comm: communicator handle - not used
c     err: returned status - not used
c**********************************************************
      implicit none
#include "t3e.h"
      integer sem1,semwait,semset
      common/semaphores/sem1,semwait,semset
c     *** Arguments 
      integer msg(*)
      integer count,datatype,dest,tag,comm,err

#include "parallel.h"
#ifdef MPI_VERBOSE
      write (6,*) 'PE',mytaskid,' MPI_SEND',dest,count
#endif

c     Check if enough buffer space for msg
      if (count .GT. T3E_BUF_SIZE) then
         write (6, '(/,a,/,a,i10,/,a)') 
     .        'ERROR, T3E implementation. Not enough buffer ',
     .        'space in SEND. MSG size is: ', count,
     .        'Increase T3E_BUF_SIZE in t3e.h'
         call mexit(6,1)
      endif
c     JV Wait on semaphore until receiver sets
      if(sem1.EQ.semwait)call shmem_wait(sem1,semwait)
      sem1 = semwait
c     Put msg into nbuf on PE dest
      call shmem_fence
      call shmem_put(nbuf,msg,count,dest)
c     Set semaphore on PE dest
      call shmem_fence
      call shmem_put(sem1,semset,1,dest) 
      return
      end


c************************************************************
c               MPI_BCAST
c************************************************************
      subroutine MPI_BCAST(msg,count,datatype,root,comm,err)
c************************************************************
c     Broadcast the array 'msg' to all nodes in the configuration
c     msg type can be anything
c     datatype: type of data being broadcast
c     root: the node which will do the broadcasting
c     group: integer identifier of group to broadcast to -not used
c     count: Number of datatype elements in msg
c     (NOTE: T3E ints, reals, and doubles are all the same size)
c     shmem routines have no implicit synchronization, thus 
c     external barriers are required.
c************************************************************
      implicit none
#include "t3e.h"
#include "mpif.h"
#include "parallel.h"
      integer count,datatype,root,comm,err,shmem_n_pes,shmem_my_pe
      dimension msg(*)
      real msg,tim0,tim1,tim,test
      integer i,j,k,count_1,count_2,n_start
      
      n_start=1
      count_1 = count
      count_2=0

#ifdef MPI_VERBOSE
      if(mytaskid.eq.0) write(6,*)
     $     'MPI_BCAST',count,' type ',datatype
#endif
#  ifdef GP_BCAST1
            call MPI_BARRIER(comm,1)
            call shmem_broadcast(msg,msg,count,
     $           mpes(comm,1),0,mpes(comm,2)
     $           ,bcastsync(1,bcast_cnt(comm),comm) )
            bcast_cnt(comm) = 
     $           mod(bcast_cnt(comm),T3E_BCSYNC_SZ) + 1
#  else
            call MPI_BARRIER(comm,1)
            call shmem_broadcast(msg,msg,count,
     $           root,0,0,numtasks
     $           ,bcastsync(1,bcast_cnt) )
            bcast_cnt = mod(bcast_cnt,T3E_BCSYNC_SZ) + 1
#  endif
      call MPI_BARRIER(MPI_COMM_WORLD,1)

      return 
      end


#ifdef NO_SHMEMFENCE
      subroutine shmem_fence()
      return
      end
#endif
      

c************************************************************
c               MPI_REDUCE ORIGINAL JV broken for t3e
c************************************************************
      subroutine MPI_REDUCE_JV(from,to,count,datatype,op,root,comm,err) 
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array. Reduction operation is
c     double precision add only. Result is placed _on all PEs_.
c     (NOTE: T3E ints, reals, and doubles are all the same size)
c     shmem routines have no implicit synchronization, thus 
c     external barriers are required.
c******************************************************************
      implicit none
#include "t3e.h"
#include "mpif.h"
#include "parallel.h"

      integer count,datatype,op,root,comm,err,shmem_n_pes,shmem_my_pe
      integer i

      dimension from(*),to(*)
      real from,to

      if (wrkdata_size .lt. (count/2)+1) then
         write (6, '(/,a,/,a,i10,/,a)') 
     .        'ERROR, T3E implementation.  Trying to reduce too',
     .        'large a dataset.  Size is: ', count,
     .        'Increase wrkdata_size in t3e.h'
         call mexit(6,1)
      endif

      call MPI_BARRIER(MPI_COMM_WORLD,1)

      call shmem_real8_sum_to_all(to,from,count,root,0,numtasks,
     +                           wrkdata,datasync(1,data_cnt))
      data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1

      call MPI_BARRIER(MPI_COMM_WORLD,1)
      return
      end

c************************************************************
c               MPI_ALLREDUCE ORIGINAL JV broken for t3e
c************************************************************
      subroutine MPI_ALLREDUCE_JV(from,to,count,datatype,op,comm,err)
c************************************************************
c     Reduce array 'from' across all nodes and place result
c     on all nodes in 'to'.
c************************************************************
      implicit none
#include "t3e.h"
#include "mpif.h"
#include "parallel.h"

      integer count,datatype,op,comm,err,shmem_n_pes,shmem_my_pe
      integer i
      real from,to
      dimension from(*),to(*)

      if (wrkdata_size .lt. (count/2)+1) then
         
         write (6, '(/,a,/,a,i10,/,a)') 
     .        'ERROR, T3E implementation.  Trying to reduce too',
     .        'large a dataset.  Size is: ', count,
     .        'Increase wrkdata_size in t3e.h'
         call mexit(6,1)
      endif

      call MPI_BARRIER(MPI_COMM_WORLD,1)

      if(    (datatype.EQ.MPI_DOUBLE_PRECISION)
     +     .OR.(datatype.EQ.MPI_REAL)  )  then
         call shmem_real8_sum_to_all(to,from,count,0,0,numtasks,
     +        wrkdata,datasync(1,data_cnt))
         data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1
      else
         call shmem_int8_sum_to_all(to,from,count,0,0,shmem_n_pes(),
     +        wrkdata,datasync(1,data_cnt))
         data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1
      endif

      call MPI_BARRIER(MPI_COMM_WORLD,1)

      return
      end

c******************************************************************
c                       MPI_REDUCE
c******************************************************************
      subroutine MPI_REDUCE(from,to,count,datatype,op,root,comm,err) 
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array. Reduction operation is
c     double precision add only. Result is placed _on all PEs_.
c     (NOTE: T3E ints, reals, and doubles are all the same size)
c     shmem routines have no implicit synchronization, thus 
c     external barriers are required.
c******************************************************************
      implicit none
#include "t3e.h"
#include "mpif.h"
#include "parallel.h"

      integer count,datatype,op,root,comm,err,shmem_n_pes,shmem_my_pe
      integer i,j,n_msg, n_left,n_start, n_piece,n,count_1
      dimension from(*),to(*)
      real from,to

      if (wrkdata_size .lt. (count/2)+1) then
         write (6, '(/,a,/,a,i10,/,a)') 
     .        'ERROR, T3E implementation.  Trying to reduce too',
     .        'large a dataset.  Size is: ', count,
     .        'Increase wrkdata_size in t3e.h'
         call mexit(6,1)
      endif

      call MPI_BARRIER(MPI_COMM_WORLD,1)

      call shmem_real8_sum_to_all(to,from,count,root,0,numtasks,
     +                           wrkdata,datasync(1,data_cnt))
      data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1
      call MPI_BARRIER(MPI_COMM_WORLD,1)

 999   return
      end
 
c************************************************************
c               MPI_ALLREDUCE
c************************************************************
      subroutine MPI_ALLREDUCE(from,to,count,datatype,op,comm,err)
c************************************************************
c     Reduce array 'from' across all nodes and place result
c     on all nodes in 'to'.
c************************************************************
      implicit none
#include "t3e.h"
#include "mpif.h"
#include "parallel.h"

      integer count,datatype,op,comm,err,shmem_n_pes,shmem_my_pe
      integer i,j,n_msg, n_left,n_start, n_piece,n,count_1,count_2
      real from,to
      dimension from(*),to(*)

      if (wrkdata_size .lt. (count/2)+1) then
         
         write (6, '(/,a,/,a,i10,/,a)') 
     .        'ERROR, T3E implementation.  Trying to reduce too',
     .        'large a dataset.  Size is: ', count,
     .        'Increase wrkdata_size in t3e.h'
         call mexit(6,1)
      endif


      if((datatype.EQ.MPI_DOUBLE_PRECISION)
     $     .OR.(datatype.EQ.MPI_REAL)  )then

#ifdef MPI_VERBOSE
         if(mytaskid.eq.0) write(6,*) "PE0 REAL ALL_REDUCE",count
# endif

         call MPI_BARRIER(MPI_COMM_WORLD,1)

         call shmem_real8_sum_to_all(to,from,count,0,0,numtasks,
     +        wrkdata,datasync(1,data_cnt))
         data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1
 999   continue


      else
c      ==============================================================
c                    Integer all_reduce
c      ==============================================================

          call shmem_int8_sum_to_all(to,from,count,0,0,shmem_n_pes(),
     +         wrkdata,datasync(1,data_cnt))
          data_cnt = mod(data_cnt,T3E_DATASYNC_SZ)+1
       endif
       
       call MPI_BARRIER(MPI_COMM_WORLD,1)
       
       return
       end


c************************************************************
c            MPI_GROUP_INCL
c
c            MPI_GROUP_INCL(GROUP, N, RANKS, NEWGROUP, IERROR)
c            INTEGER GROUP, N, RANKS(*), NEWGROUP, IERROR 
c************************************************************
      subroutine MPI_GROUP_INCL(GROUP, N, RANKS, NEWGROUP, IERROR)
      implicit none
                  
      INTEGER GROUP, N, RANKS(*), NEWGROUP, IERROR 
#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      integer i
      
      do i=0,TOTAL_MPI_COMMUNICATORS
         if(mpi_numgp_active(i).eq. MPI_GROUP_NULL)then
            newgroup = i
            mpi_numgp_active(i)=1
            goto 10
         endif
      enddo
 10   mpi_gpsize(newgroup)=n
c       The ranks are 
      do i=0,n-1
         mpi_gpranks(i,newgroup)=ranks(i+1)
      enddo
      
      return
      end
      
c************************************************************
c      MPI_COMM_CREATE
c      
c      MPI_COMM_CREATE(COMM, GROUP, NEWCOMM, IERROR)
c      INTEGER COMM, GROUP, NEWCOMM, IERROR 
c************************************************************
      subroutine MPI_COMM_CREATE(COMM, GROUP, NEWCOMM, IERROR)
      implicit none
#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER COMM, GROUP, NEWCOMM, IERROR 
      integer i,j,nt,n,shmem_my_pe
      
      call mpi_group_size(group,n,ierror)
      nt = shmem_my_pe()
      do i=0,n-1
         if(nt.eq.mpi_gpranks(i,group))then
            do j=0,TOTAL_MPI_COMMUNICATORS
               if(mpi_ncomm_active(j).eq. MPI_COMM_NULL)then
                  newcomm = j
                  mpi_ncomm_active(j)=group
                  mpi_gpcommnum(j)=group
                  mpi_cmsize(j)=n
                  mpi_cmrank0(j)=mpi_gpranks(0,group)
                  mpes(newcomm,1)=mpi_cmrank0(j)
                  mpes(newcomm,2)=n
                  return
               endif
            enddo
         endif
      enddo
      newcomm=MPI_COMM_NULL
      return
      end

c************************************************************
c      MPI_GROUP_FREE
c
c      MPI_GROUP_FREE(GROUP, IERROR)
c      INTEGER GROUP, IERROR 
c************************************************************
      subroutine MPI_GROUP_FREE(GROUP, IERROR)
      implicit none

#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER GROUP, IERROR 
      
      mpi_numgp_active(group)= MPI_GROUP_NULL
      return
      end
c************************************************************
c      MPI_COMM_GROUP
c
c      MPI_COMM_GROUP(COMM,GROUP, IERROR)
c      INTEGER COMM,GROUP, IERROR 
c************************************************************
      subroutine MPI_COMM_GROUP(COMM,GROUP, IERROR)
      implicit none

#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER COMM,GROUP, IERROR 
      
      group=mpi_ncomm_active(comm)
   
      return
      end

c************************************************************
c      MPI_COMM_FREE
c
c      MPI_COMM_FREE(COMM, IERROR)
c      INTEGER COMM, IERROR 
c************************************************************
      subroutine MPI_COMM_FREE(COMM, IERROR)
      implicit none

#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER COMM, IERROR 
      
      mpi_ncomm_active(comm)= MPI_COMM_NULL
      mpi_gpcommnum(comm)=MPI_GROUP_NULL
      return
      end

c************************************************************
c      MPI_COMM_SIZE
c
c      MPI_COMM_SIZE(COMM, SIZE, IERROR)
c      INTEGER COMM, SIZE, IERROR 
c************************************************************
      subroutine MPI_COMM_SIZE(COMM, SIZE, IERROR)
      implicit none

#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER COMM, SIZE, IERROR ,group
      
      group = mpi_gpcommnum(comm)
      size = mpi_gpsize(group)

      return
      end
c************************************************************
c      MPI_GROUP_SIZE
c
c      MPI_GROUP_SIZE(GROUP, SIZE, IERROR)
c      INTEGER GROUP, SIZE, IERROR 
c************************************************************
      subroutine MPI_GROUP_SIZE(GROUP, SIZE, IERROR)
      implicit none

#     include "parallel.h"
#     include "mpif.h"
#     include "t3e.h"
      INTEGER GROUP, SIZE, IERROR 
      size = mpi_gpsize(group)
      return
      end

