c************************************************************
c
c     This is an interface for MPI calls using TCGMSG 
c     for message passing.
c     
c     The file is based on the mpi/template.f file from
c     James J. Vincent.
c    
c     This interface was written by Thomas Huber
c     Ludwig Maximilian Universitaet Muenchen, Germany
c     email: thuber@physik.tu-muenchen.de
c
c
c     To compile sander with TCGMSG/MPI-parallel options
c     add -DMPI -DTCGMSG to the MACHINEFLAGS variable.
c     If TCGMSG is executed on different architectures
c     across one network, add -DXDR. (Not tested!)
c     Then the the libtcgmsg.a file has to be added to the
c     LOADLIB variable.
c
c     See the TCGMSG manual for executing parallel jobs
c     on a particular machine.
c
c     The original MPI code was written and contributed by:
c
c     James J. Vincent
c     Kenneth M. Merz
c     Department of Chemistry
c     Pennsylvania State University
c     July, 1994
c
c     tec3: added comments, more error checking, mpi_abort, 
c     wrappers for the reduce calls and fixed typing in the
c     recv calls.
c     Note that "tcgmsg.h" is equivalent to the standard 
c     TCGMSG include file "msgtypesf.h"
c     Note also that TCGMSG has a nice feature for reporting
c     all messages sent and received.  To enable this, add
c     a call to setdbg(1) somewhere in the code; a good place
c     is probably the subroutine startup defined in parallel.f
c
c***********************************************************

      subroutine MPI_BARRIER(group)
c***********************************************************
c     Synchronize all tasks
**********************************************************
#include "mpif.h"      
#include "parallel.h"
#include "tcgmsg.h"      
      integer group
      
      call synch(1000)

      return
      end

      subroutine MPI_ABORT(ierr)
c************************************************************
c     Abort out and kill all processes as necessary
c************************************************************
#include "tcgmsg.h"
      integer ierr

      call parerr(ierr)

      end

      subroutine MPI_FINALIZE(ierr)
c************************************************************
c     Execute any exit code specific to parallel machine
c************************************************************
#include "mpif.h"
#include "parallel.h"
#include "tcgmsg.h"
      integer ierr

      call pend

      end
      
      subroutine MPI_INIT(ierr)
c*************************************************************
c     Enrolls caller in parallel machine.
c     Must set number of tasks and own PE number.
c*************************************************************
#include "mpif.h"
#include "parallel.h"
#include "tcgmsg.h"
      integer ierr

      call pbeginf

      numtasks = nnodes()
      mytaskid = nodeid()
      
      if ( mytaskid .eq. 0 ) then
         master = .true.
      else
         master = .false.
      endif

      return
      end

      subroutine MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
c************************************************************
c     Get my PE number (rank) 
c************************************************************
#include "parallel.h"
#include "tcgmsg.h"
      integer rank,ierr

      rank = mytaskid

      return
      end


      subroutine MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)
c************************************************************
c     Get size of machine (number of PE's)
c************************************************************
#include "parallel.h"
#include "tcgmsg.h"
      integer size,ierr

      size = numtasks

      return
      end

      subroutine MPI_RECV(msg,count,datatype,source,tag,comm,ierr)
c***********************************************************
c     Receive a message 
c     msg: buffer to put message in
c     count : number of datatype elements
c     datatype: type of data
c     source: PE number of sender
c     tag: message tag
c     comm: group comm id (not used)
c     status: returned status of message receive
c**********************************************************
#include "mpif.h"
#include "parallel.h"
#include "tcgmsg.h"
c     *** Arguments 
      dimension msg(*)
      double precision msg
      integer count,datatype,source,tag,comm,ierr
      integer nbytes, lenmes, nodefrom, type

c
c     ...determine the number of bytes per word
c
      if ( datatype .eq. MPI_INTEGER2 ) then
         nbytes = 2
      else if ( (datatype .eq. MPI_INTEGER) .or.
     .          (datatype .eq. MPI_REAL) ) then
# ifdef MACH64
         nbytes = 4
# else
         nbytes = 4
# endif
      else if ( datatype .eq. MPI_DOUBLE_PRECISION) then
         nbytes = 8
      endif

      type = MSGCHR

# ifdef XDR
c
c     Where possible, TCGMSG attempts to use XDR communication over
c     sockets to optimize communication.  This is enabled by specifying
c     a higher order bit in the "type" parameter which can be 
c     accomplished by adding in one of the type descriptors present in
c     tcgmsg.h (msgtypesf.h), currently one of MSGCHR (characters), 
c     MSGINT (integers), or MSGDBL (double precision).
c
      if ( datatype .eq. MPI_INTEGER ) then
         type = MSGINT
      else if ( datatype .eq. MPI_DOUBLE_PRECISION ) then
         type = MSGDBL
      endif
# endif
c
c     ...set the actual number of bytes
c
      nbytes = count * nbytes
c
c     ...make the lower order bits of type match the tag
c
      type = type + tag
c
c     ...the current implementation uses type = MSGCHR (for xdr)
c     by default
c
      type = MSGCHR

#ifdef TCG_ASYNCH
      call rcv( type, msg, nbytes, lenmes, source, nodefrom, 0 )
#else
      call rcv( type, msg, nbytes, lenmes, source, nodefrom, 1 )
#endif

      if ( nbytes .ne. lenmes ) then
         write(6, '(''ERROR in MPI_RECV(): message length wrong'')')
         write(6, '(a,i8,a,i8)') 
     .        'Expected ', nbytes, 
     .        ' bytes but only received ', lenmes
         call mexit(6,1)
      endif

      if ( (numtasks .lt. nodefrom) .or.
     +     (0 .gt. nodefrom ) ) then
         write(6, '(a,i5,a)') 'ERROR in MPI_RECV(): node ', nodefrom,
     .        ' is out of bounds!'
         call mexit(6,1)
      endif

      if ( (source .ne. -1) .and. (source .ne. nodefrom) ) then
         write(6, '(a,i5,a,i5)') 
     .        'ERROR in MPI_RECV(): message from wrong node?',
     .        source, ' does not match ', nodefrom
         call mexit(6,1)
      endif

      return
      end

      subroutine MPI_SEND(msg,count,datatype,dest,tag,comm,ierr)
c***********************************************************
c     Send a message 
c     msg: buffer message is in
c     count: number of datatype elements in message
c     datatype MPI datatype of message elements
c     dest: PE number of destination process (0->n-1)
c     tag: message type
c     comm: group identifier for communication (not used yet)
c**********************************************************
#include "mpif.h"
#include "tcgmsg.h"
      dimension msg(*)
      double precision msg

      integer count, datatype, dest, tag, comm, ierr
      integer nbytes, type
c
c     ...determine the number of bytes per word
c
      if ( datatype .eq. MPI_INTEGER2 ) then
         nbytes = 2
      else if ( (datatype .eq. MPI_INTEGER) .or.
     .          (datatype .eq. MPI_REAL) ) then
#ifdef MACH64
         nbytes = 4
#else
         nbytes = 4
#endif
      else if ( datatype .eq. MPI_DOUBLE_PRECISION) then
         nbytes = 8
      endif

      type = MSGCHR

#ifdef XDR
c
c     Where possible, TCGMSG attempts to use XDR communication over
c     sockets to optimize communication.  This is enabled by specifying
c     a higher order bit in the "type" parameter which can be 
c     accomplished by adding in one of the type descriptors present in
c     tcgmsg.h (msgtypesf.h), currently one of MSGCHR (characters), 
c     MSGINT (integers), or MSGDBL (double precision).
c
      if ( datatype .eq. MPI_INTEGER ) then
         type = MSGINT
      else if ( datatype .eq. MPI_DOUBLE_PRECISION ) then
         type = MSGDBL
      endif
#endif

c
c     ...set the actual number of bytes
c
      nbytes = count * nbytes
c
c     ...make the lower order bits of type match the tag
c
      type = type + tag
c
c     ...the current implementation uses type = MSGCHR (for xdr)
c     by default
c
      type = MSGCHR


#ifdef TCG_ASYNCH
      call snd( type, msg, nbytes, dest, 0 )
#else
      call snd( type, msg, nbytes, dest, 1 )
#endif

      return
      end

      subroutine MPI_BCAST(msg,count,datatype,root,comm,ierr)
c************************************************************
c     Broadcast the array 'msg' to all nodes in the configuration
c     msg type can be anything
c     datatype: type of data being broadcast
c     root: the node which will do the broadcasting
c     group: integer identifier of group to broadcast to -not used
c     count: Number of datatype elements in msg
c************************************************************
#include "mpif.h"
#include "tcgmsg.h"

      integer count,datatype,root,comm,ierr,stype
      
      integer nbytes, type

c
c     ...determine the number of bytes per word
c
      if ( datatype .eq. MPI_INTEGER2 ) then
         nbytes = 2
      else if ( (datatype .eq. MPI_INTEGER) .or.
     .          (datatype .eq. MPI_REAL) ) then
#ifdef MACH64
         nbytes = 4
#else
         nbytes = 4
#endif
      else if ( datatype .eq. MPI_DOUBLE_PRECISION) then
         nbytes = 8
      endif

      type = MSGCHR

#ifdef XDR
c
c     Where possible, TCGMSG attempts to use XDR communication over
c     sockets to optimize communication.  This is enabled by specifying
c     a higher order bit in the "type" parameter which can be 
c     accomplished by adding in one of the type descriptors present in
c     tcgmsg.h (msgtypesf.h), currently one of MSGCHR (characters), 
c     MSGINT (integers), or MSGDBL (double precision).
c
      if ( datatype .eq. MPI_INTEGER ) then
         type = MSGINT
      else if ( datatype .eq. MPI_DOUBLE_PRECISION ) then
         type = MSGDBL
      endif
#endif
c
c     ...set the actual number of bytes
c
      nbytes = count * nbytes
c
c     ...make the lower order bits of type match the tag
c
      type = type + tag

#ifndef XDR
      type = MSGCHR
#endif

      call brdcst( type, msg, nbytes, root )

      return 
      end
      

      subroutine MPI_REDUCE(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Wrapper to dispatch on datatype, 
c     based on JV's code in pvm.f (tec3)
c******************************************************************
#include "mpif.h"
      integer datatype
c
      if (datatype.eq.MPI_INTEGER) then
         call MPI_REDUCE_I(from,to,count,datatype,op,root,comm,ierr)
      else if (datatype.eq.MPI_DOUBLE_PRECISION) then
         call MPI_REDUCE_D(from,to,count,datatype,op,root,comm,ierr)
      else
         write(6,'(/5x,a,i5)')
     .        'ERROR in MPI_REDUCE(): unsupported datatype is ',
     .        datatype
         call mexit(6,1)
      endif
      return
      end


      subroutine MPI_REDUCE_I(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array.  
c     INTEGER VERSION
c*****************************************************************
#include "mpif.h"
#include "parallel.h"
#include "tcgmsg.h"

c     *** Arguments ****
      integer count,datatype,op,root,comm,ierr
c     This nodes portion of the array to reduce
      dimension from(*)
      integer from
c     Array to store result in
      dimension to(*)
      integer to

      integer i

      do i=1, count
          to(i)=from(i) 
      enddo
      if (op .eq. MPI_SUM) then

         call igop( 1001, to, count, '+' )

      else
         write(6,'(/,a,i5)') 
     .        'Unknown operator in MPI_REDUCE is ', op
         write(6,'(a)') 'Only MPI_SUM is supported currently'
         call mexit(6,1)
      endif

      return
      end

      subroutine MPI_REDUCE_D(from,to,count,datatype,op,root,comm,ierr)
c******************************************************************
c     Perform reduction operation on the array 'from' across all
c     nodes and place result in 'to' array.
c     DOUBLE PRECISION VERSION
c*****************************************************************
#include "mpif.h"
#include "parallel.h"
#include "tcgmsg.h"

c     *** Arguments ****
      integer count,datatype,op,root,comm,ierr
c     This nodes portion of the array to reduce
      dimension from(*)
      double precision from
c     Array to store result in
      dimension to(*)
      double precision to

      integer i

      do i=1, count
          to(i)=from(i) 
      enddo
      if (op .eq. MPI_SUM) then

         call dgop( 1001, to, count, '+' )

      else
         write(6,'(/,a,i5)') 
     .        'Unknown operator in MPI_REDUCE is ', op
         write(6,'(a)') 'Only MPI_SUM is supported currently'
         call mexit(6,1)
      endif

      return
      end

      subroutine MPI_ALLREDUCE(from,to,count,datatype,op,comm,ierr)
c******************************************************************
c     Wrapper to dispatch on datatype, 
c     based on JV's code in pvm.f (tec3)
c******************************************************************
#include "mpif.h"
      integer datatype
c
      if (datatype.eq.MPI_INTEGER) then
         call MPI_ALLREDUCE_I(from,to,count,datatype,op,comm,ierr)
      else if (datatype.eq.MPI_DOUBLE_PRECISION) then
         call MPI_REDUCE_D(from,to,count,datatype,op,comm,ierr)
      else
         write(6,'(/5x,a,i5)')
     .        'ERROR in MPI_ALLREDUCE(): unsupported datatype is ',
     .        datatype
         call mexit(6,1)
      endif
      return
      end

      subroutine MPI_ALLREDUCE_I(from,to,count,datatype,op,comm,ierr)
c************************************************************
c     Perform reduction operation on the array 'from' from all
c     nodes then broadcast the resulting array back to all nodes
c     and place on 'to'. 
c     Only add supported now, but others are easily added
c     INTEGER VERSION
c************************************************************
#include "tcgmsg.h"
#include "parallel.h"
#include "mpif.h"

c     *** Argumenst
      dimension from(*),to(*)
      integer to,from
      integer count,datatype,op,comm,ierr
      
c     Use node 0 as the central gatherer (debug)
c     call MPI_REDUCE(from,to,count,datatype,op,0,comm,ierr)
c     call MPI_BCAST(to,count,datatype,0,comm,ierr)

      integer i

      do i=1, count
          to(i)=from(i) 
      enddo

      if (op .eq. MPI_SUM) then

         call igop( 1002, to, count, '+' )

      else
         write(6,'(/,a,i5)') 'Unknown operator in MPI_ALLREDUCE is ',
     .        op
         write(6,'(a)') 'Only MPI_SUM is supported currently'
         call mexit(6,1)
      endif
      return
      end


      subroutine MPI_ALLREDUCE_D(from,to,count,datatype,op,comm,ierr)
c************************************************************
c     Perform reduction operation on the array 'from' from all
c     nodes then broadcast the resulting array back to all nodes
c     and place on 'to'. 
c     Only add supported now, but others are easily added
c     DOUBLE PRECISION VERSION
c************************************************************
#include "tcgmsg.h"
#include "parallel.h"
#include "mpif.h"

c     *** Argumenst
      dimension from(*),to(*)
      double precision to,from
      integer count,datatype,op,comm,ierr
      
c     DEBUG: Use node 0 as the central gatherer
c     call MPI_REDUCE(from,to,count,datatype,op,0,comm,ierr)
c     call MPI_BCAST(to,count,datatype,0,comm,ierr)
c     return

      integer i

      do i=1, count
          to(i)=from(i) 
      enddo

      if (op .eq. MPI_SUM) then

         call dgop( 1002, to, count, '+' )

      else
         write(6,'(/,a,i5)') 'Unknown operator in MPI_ALLREDUCE is ',
     .        op
         write(6,'(a)') 'Only MPI_SUM is supported currently'
         call mexit(6,1)
      endif
      return
      end

      subroutine MPI_CONCAT(outmsg,inmsg,count,datatype,comm,ierr)
c*********************************************************
c     Concatenate outmsg from each process into inmsg by
c     order of PE #. Inmsg is (nhosts)*outmsg in length
c     outmsg: local message to get concatenated
c     inmsg: copy of all messages, placed back on all tasks
c     size: size of OUTMSG in BYTES
c     group: integer identifying group of tasks (not used)
c     MSGTYPE 102 used here
c     !!!!outmsg gets destroyed in the process!!!
c************************************************************
C	#include <fpvm3.h>
#include "parallel.h"     
#include "tcgmsg.h"

      common/locpvm/ntids(0:511),myparent

c     *** Arguments
      dimension outmsg(*),inmsg(*)
      integer outmsg,inmsg
      integer count,datatype,comm,ierr

c     *** Local 
      integer i

      call MPI_BARRIER(MPI_COM_WORLD,1)

c     Everyone send their data to 0 (including 0)
      call MPI_SEND(outmsg,count,datatype,0,102,comm,ierr)

c     PE # 0 does the work
      if(mytaskid .eq. 0) then      
c     Make sure to recv messages in order
         do i = 0,numtasks-1
          call MPI_RECV(inmsg(i*count+1),count,datatype,i,102,comm,ierr)
         enddo
      endif
c     Broadcast concatenated msg back to all
      call MPI_BCAST(inmsg,count*numtasks,datatype,0,comm,ierr)
      return
      end


