#! /bin/csh -f
#
#-----------------------------------------------------------+
#                                                           |
# Unix shell script for parsing command line, setting up    |
# file links, and running SPASMS                            |
#                                                           |
# flags are:                                                |
#     files:    purpose     	default extension	    |
#	-i	control input   .in			    |
#	-o	output log	.out			    |
#	-p	parm topology	.top			    |
#	-c	start coords	.inp			    |
#	-r	final coords	.rst			    |
#	-d	dump 		.dmp			    |
#       -s      summary 	.sum			    |
#	-e	energy log	.en			    |
#	-v	velocity log	.vel			    |
#	-x                                                  |
#    or -t	coordinate log	.crd			    |
#     -ref                                                  |
#    or -z	reference crds	.ref			    |
#	-dd	dihedral driver	NO DEFAULT		    |
#	-nmr	flatwell input	NO DEFAULT		    |
#    others:						    |
#       -ri     input root name default  = SPIN             |
#       -ro     output rootname default  = SPOT             |
#       -ei     input extension default  = .fmt             |
#       -eo     output extensin default  = .fin             |
#	-T	arg is test version of SPASMS executable    |
#		(allows testing new versions before install)|
#                                                           |
#                                                           |
# Erik-Robert Evensen                                       |
# 11 May 1990                                               |
#                                                           |
#-----------------------------------------------------------+

set SPASMS = spasms
set lerror = 0
#limit stacksize 16384	# this may not work with all C shells

#
#  the use of symbolic links for output files doesn't work w/
# 	HP, so the script is dumbed down
#
set HPBUG = 1

#
#  See if system supports symbolic links
#
if ( `ln -s /tmp/ln1.$$ /tmp/ln2.$$` ) then
	set SYMLINKS = 0
else
	set SYMLINKS = 1
	\rm -f /tmp/ln2.$$
endif

#
# remove any links or default files which get opened as type 'new'
#
if (-e MDOUT)   \rm MDOUT
if (-e MDEN)    \rm MDEN
if (-e MDVEL)   \rm MDVEL
if (-e RESTRT)  \rm RESTRT
if (-e MDDUMP)  \rm MDDUMP
if (-e MDCRD)   \rm MDCRD
if (-e MDSUM)   \rm MDSUM

#
# check how many arguments
#
if ($#argv == 0) then
	echo "No arguments given -- using default names"
	time $SPASMS
	exit
endif

#
#  Process args, creating links or copying input files.
#
#	Input files
#		if SYMLINKS:    ln -s file.in MDIN
#			creates MDIN link for spasms to open, which
#			will result in opening file.in
#		else 		cp file.in MDIN
#			spasms will open a copy of MDIN, and the
#			original file.in will safely retain its name
#	Output files:    <do nothing>
#		when spasms is finished, move MDOUT to file.out
#
while ($#argv)
	switch ($1)
		case -i:				# name of input file
			shift
			set in = $1
			if ($1:e == "") set in = $1.in
			if (! -e $in) then
				echo "Input file ($in) missing"
				set lerror = 1
			else
				if (-e MDIN) \rm MDIN
				if ( $SYMLINKS ) then
					ln -s $in MDIN
				else
					cp    $in MDIN
				endif
			endif
			shift
			breaksw
		case -o:				# name of output file
			shift
			set out = $1
			if ($1:e == "") set out = $1.out
			if (-e $out) then
				echo "Output file ($out) exists"
				set lerror = 1
			endif
			shift
			breaksw
		case -p:				# name of parm top file
			shift
			set prmtop = $1
			if ($1:e == "") set prmtop = $1.top
			if (! -e $prmtop) then
				echo "Parameter file ($prmtop) missing"
				set lerror = 1
			else
				if (-e PRMTOP) \rm PRMTOP
				if ( $SYMLINKS ) ln -s $prmtop PRMTOP
				else             cp    $prmtop PRMTOP
			endif
			shift
			breaksw
		case -c:				# name of inpcrd file
			shift
			set inpcrd = $1
			if ($1:e == "") set inpcrd = $1.inp
			if (! -e $inpcrd) then
				echo "Input crd ($inpcrd) file missing"
				set lerror = 1
			else
				if (-e INPCRD) \rm INPCRD
				if ( $SYMLINKS ) ln -s $inpcrd INPCRD
				else             cp    $inpcrd INPCRD
			endif
			shift
			breaksw
		case -r:				# name of restrt file
			shift
			set restrt = $1
			if ($1:e == "") set restrt = $1.rst
			if (-e restrt) then
				echo "Restart file ($restrt) exists"
				set lerror = 1
			endif
			shift
			breaksw
		case -ref:
		case -z:				# name of ref crd file
			shift
			set refc = $1
			if ($1:e == "") set refc = $1.ref
			if (! -e $refc) then
				echo "Reference coord file ($refc) missing"
				set lerror = 1
			else
				if (-e REFC) \rm REFC
				if ( $SYMLINKS ) ln -s $refc REFC
				else		 cp    $refc REFC
			endif
			shift
			breaksw
		case -e:				# name of energy file
			shift
			set mden = $1
			if ($1:e == "") set mden = $1.en
			if (-e $mden) then
				echo "Energy file ($mden) exists"
				set lerror = 1
			endif
			shift
			breaksw
		case -v:				# name of velocity file
			shift
			set mdvel = $1
			if ($1:e == "") set mdvel = $1.vel
			if (-e $mdvel) then
				echo "Velocity file ($mdvel) exists"
				set lerror = 1
			endif
			shift
			breaksw
		case -x:
		case -t:				# name of coord file
			shift
			set mdcrd = $1
			if ($1:e == "") set mdcrd = $1.crd
			if (-e $mdcrd) then
				echo "Coordinate file ($mdcrd) exists"
				set lerror = 1
			endif
			shift
			breaksw
		case -d:				# name of dump file
			shift
			set mddump = $1
			if ($1:e == "") set mddump = $1.dmp
			shift
			breaksw
		case -s:				# name of summary file
			shift
			set mdsum = $1
			if ($1:e == "") set mdsum = $1.sum
			shift
			breaksw
		case -ei:			# extension for multiple files
			shift
			set exteni = $1
			shift
			breaksw
		case -eo:			# extension for multiple files
			shift
			set exteno = $1
			shift
			breaksw
		case -ro:			# root name for multiple files
			shift
			set rootot = $1
			shift
			breaksw
		case -ri:			# root name for multiple files
			shift
			set rootin = $1
			shift
			breaksw
		case -dd:			# dihedral driver input file
			shift
			set mddrv = $1
			ln -s $mddrv MDDRV
			shift
			breaksw
		case -nmr:			# dihedral driver input file
			shift
			set mdnmr = $1
			if ( -e MDNMR ) \rm MDNMR
			if ( $SYMLINKS ) then
				ln -s $mdnmr MDNMR
			else 
				cp $mdnmr MDNMR
			endif
			shift
			breaksw
		case -T:			# name of test executable
			shift
			set SPASMS = $1
			if (! -x $SPASMS) then
				echo "Test executable ($SPASMS) not found"
				set lerror = 1
			endif
			shift
			breaksw
		default:				# unrecognized flag
			echo "Flag $argv is not recognized - exiting"
			set lerror = 1
			exit
		
	endsw
end

if ($lerror == 1) then
	echo "One or more errors detected with file names - exiting"
	exit(1)
endif

if ($?in) then
	echo "Input file is $in"
endif
if ($?out) then
	echo "Output file is $out"
endif

onintr fixfiles
time $SPASMS				# run SPASMS

fixfiles:
#
#  'rename' output files; determine success based on presence
#	of timing info at end of .out
#
        if ($?out) then
                mv MDOUT $out
		set success = ` tail -5 $out | egrep "Which Equals" `
	else
		set success = ` tail -5 MDOUT | egrep "Which Equals" `
        endif
        if ($?en) then
                mv MDEN $en
        endif
        if ($?vel) then
                mv MDVEL $vel
        endif
        if ($?restrt) then
                mv RESTRT $restrt
        endif
        if ($?dump) then
		touch $dump
                cat MDDUMP >>  $dump
        endif
        if ($?crd) then
                mv MDCRD $crd
        endif
#
#  remove any links or copied files
#
	if ($?in)	\rm -f MDIN
	if ($?out)	\rm -f MDOUT
	if ($?prmtop)	\rm -f PRMTOP
	if ($?inpcrd)	\rm -f INPCRD
	if ($?mden)	\rm -f MDEN
	if ($?mdvel)	\rm -f MDVEL
	if ($?refc)	\rm -f REFC
	if ($?restrt)	\rm -f RESTRT
	if ($?mddump)	\rm -f MDDUMP
	if ($?mdcrd)	\rm -f MDCRD
	if ($?mdsum)	\rm -f MDSUM
	if ($?mddrv)	\rm -f MDDRV
	if ($?mdnmr)	\rm -f MDNMR


if ( "$success" != "" ) then
	echo No errors detected
	exit(0)
endif
echo EXECUTION PROBLEM
exit(1)




